;; Build and maintain an irc connection.
;; Copyright (C) 2001  Neal H Walfield  <neal@cs.uml.edu>
;;
;; This file is part of the Deride
;;
;; Deride is free software; you can redistribute it and/or modify
;; it under the terms of the GNU General Public License as published by
;; the Free Software Foundation; either version 2, or (at your option)
;; any later version.
;; 
;; Deride is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.
;; 
;; You should have received a copy of the GNU General Public License
;; along with the Deride; see the file COPYING.  If not, write to
;; the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

(define-module (comm irc irc))

(use-modules (ice-9 slib))
(use-modules (ice-9 regex))
(use-modules (comm irc rfc-2812))
(use-modules (forum users))

;; The end of line marker for irc servers.
(define-public irc-eol "\n\r")

;; The user's nick.
(define irc-nick "")

;; Connect to SERVER on PORT.  Return a SOCKET.
(define-public (irc-connect server port)
  (let*
      ((s (socket AF_INET SOCK_STREAM 0))
       (host (catch
	      #t
	      (lambda () (hostent:addr-list (gethostbyname server)))
	      (lambda args #f))))

    (and host
	 (catch
	  #t
	  (lambda () (connect s AF_INET (car host) port))
	  (lambda args #f))
	 s)))

;; Send DATA to SOCKET and the STDOUT
(define-public (irc-send-raw data socket)
;  (display (string-append data "\n"))
  (write-line (string-append data irc-eol) socket))

;; Login to the irc server on SOCKET using NICK and REALNAME
(define-public (irc-send-login socket nick realname)
  (set! irc-nick nick)
  (irc-send-raw (string-append "USER " nick " 0 * :" realname)
		       socket)
  (irc-send-raw (string-append "NICK " nick) socket))

;; Outstanding number of pings to the server
(define irc-outstanding-server-pings 0)

;; Ping the server.
(define (irc-ping-server socket)
  (irc-send-raw (string-append "PING " irc-server-name) socket)
  (set! irc-outstanding-server-pings (1+ irc-outstanding-server-pings)))

;; The name of the irc server we are directly connected to (as it sees
;; fit to call itself).
(define irc-server-name "unknown")

;; Turn LIST, a list of strings into a string with the elements
;; seperated by SEP and the last two elements seperated by FINALSEP.
;; E.g. (list-of-strings->string ("a" "b" "c" "d") ", " " and ")
;;       => "a, b, c and d"
(define (list-of-strings->string list sep finalsep)
  (if (eqv? list '())
      ""
      (string-append
       (if (eqv? (cdr list) '())
	   (car list)
	   (string-append (car list)
			  (if (eqv? (cddr list) '())
			      finalsep
			      sep)
			  (list-of-strings->string (cdr list)
						   sep finalsep))))))

;; Parse a packet and act on it!
(define (parse-packet data socket)
  ;; Parse an irc message RFC1459, section 2.3.1
  (define irc-message-regex
    (make-regexp
     (string-append
      "^(:(([^ !]+)!?([^ ]*)))? *"      ; The prefix: i.e. `:nick!user@host'
      "([a-zA-Z]+|[0-9]{3}) +"          ; The command
      "(([^: ]*) )?"                    ; The first middle parameter
      "(([^: ]*) )?"                    ; The second middle parameter
      "(([^: ]*) )?"                    ; The third middle parameter
      "(([^:]*):?)?"                    ; The rest of the middle parameters
      "(.*)[\n\r]+?$")))                ; Trailing

  (let ((rx (regexp-exec irc-message-regex data)))
    (if (not rx)
	(client-display (string-append "Unparsable message from server: `"
				       data "'\n"))
	(let ((prefix (or (match:substring rx 2) ""))
	      (user (or (match:substring rx 3) ""))
	      (users-host (or (match:substring rx 4) ""))
	      (command (or (match:substring rx 5) ""))
	      (first-middle (or (match:substring rx 7) ""))
	      (second-middle (or (match:substring rx 9) ""))
	      (third-middle (or (match:substring rx 11) ""))
	      (middle (or (match:substring rx 13) ""))
	      (trailing (or (match:substring rx 14) "")))

	  (cond

	   ;; Message.
	   ;; By far the most common message.  Ergo, it is at the top.
	   ((string-ci=? command "PRIVMSG")
	    (let* ((personal (if (string-ci=? first-middle irc-nick) #t #f))
		   (sep
		    (if (and
			 (>= (string-length trailing) 7)
			 (= (char->integer (string-ref trailing 0)) 1)
			 (string=? "ACTION " (substring trailing 1 8)))
			(list "" "" (substring trailing 8))
			(if personal
			    (list "*" "*" trailing)
			    (list "<" ">" trailing)))))
	      (client-display (string-append (car sep) user
					     (if personal "" first-middle)
					     (cadr sep) " " (caddr sep)
					     "\n"))))

	   ;; Notices.  The same as PRIVMSG, however, we are prohibited from
	   ;; replying to them.
	   ((string-ci=? command "NOTICE")
	    (client-display
	     (string-append
	      (if (string=? prefix "")
		  (string-append "{" irc-server-name "/" first-middle "}")
		  (string-append "<" user "/" first-middle ">"))
	      " " trailing "\n")))

	   ;; A user joined a channel
	   ((string-ci=? command "JOIN")
	    (forum-user-joined trailing user)
	    (client-display (string-append user " joined " trailing "\n")))

	   ;; A user left a channel
	   ((string-ci=? command "PART")
	    (forum-user-parted first-middle user)
	    (client-display (string-append user " has left " first-middle ": "
					   trailing "\n")))

	   ;; A user has quit.  Hm, it does not give us the channels that he
	   ;; was on.  Looks like we need a reverse map.
	   ((string-ci=? command "QUIT")
	    (client-display (string-append
			     user " has quit (was on "
			     (list-of-strings->string (forum-user-quit user)
						      ", " " and ")
			     "): " trailing "\n")))

	   ;; Same for NICK.
	   ((string-ci=? command "NICK")
	    (client-display (string-append
			     user " has changed his nick to " trailing
			     " (effects " (list-of-strings->string
					   (forum-user-schizophrenia
					    user trailing)
					   ", " " and ")
			     ").\n")))

	   ;; Just got a list of people in the channel
	   ((string-ci=? command RPL_NAMREPLY)
	    (letrec ((split (lambda (string)
			    (if (> (string-length string) 0)
				(case (string-ref string 0)
				  ((#\space #\@ #\+)
				   (split (substring string 1)))
				  (else
				   (let ((end (string-index string #\space)))
				     (forum-user-joined
				      third-middle
				      (substring string 0 end))
				     (split (substring string (1+ end))))))))))
	      (split trailing)))

	   ;; End of the list of nicks.
	   ((string=? command RPL_ENDOFNAMES)
	    (client-display (string-append
			     "Users on " second-middle ": "
			     (list-of-strings->string
			      (forum-user-list second-middle) ", " " and ")
			     "\n")))

	   ;; PING!
	   ((string-ci=? command "PING")
	    (irc-send-raw (string-append
			   "PONG " (substring data 6))
			  socket))

	   ;; PONG!
	   ((string-ci=? command "PONG")
	    (if (string-ci=? prefix irc-server-name)
		;; The server responded to a ping, mark it.  This
		;; makes the pretty reasonable assumption that the
		;; only way we will ever receive a PONG from our
		;; server is in response to a successful call to
		;; irc-ping-server (which incremented
		;; irc-outstanding-server-pings).
		(set! irc-outstanding-server-pings
		      (1- irc-outstanding-server-pings))
		(client-display (string-append "PONG from " prefix "\n"))))

	   ;; Welcome to the irc server, blah, blah, blah.  Save the
	   ;; server's name so we can ping it later; this is the easiest
	   ;; spot to get it.
	   ((or
	     (string=? command RPL_WELCOME)
	     (string=? command RPL_YOURHOST)
	     (string=? command RPL_CREATED)
	     (string=? command RPL_MYINFO))
	    (set! irc-server-name prefix)
	    (client-display (string-append "[info] " trailing "\n")))

	   ;; Server info; just echo it.
	   ((or
	     (string=? command RPL_LUSERCLIENT)
	     (string=? command RPL_LUSEROP)
	     (string=? command RPL_LUSERUNKNOWN)
	     (string=? command RPL_LUSERCHANNELS)
	     (string=? command RPL_LUSERME))
	    (client-display (string-append "[info] " second-middle
					   (if (> (string-length
						   second-middle)
						  0)
					       " "
					       "")
					   trailing "\n")))

	   ;; Try to connect to a different server.
	   ((string=? command RPL_BOUNCE)
	    (client-display (string-append "Being redirected!\n"
					   trailing "\n")))

	   ;; Channel modes.
	   ((string=? command RPL_CHANNELMODEIS)
	    (display "modes\n"))

	   ;; No topic.
	   ; What does this look like?
;	   ((string=? command RPL_NOTOPIC)
;	    (client-display data) (newline)
;	    (display "No topic for foo\n"))

	   ;; Initial topic.
	   ((string=? command RPL_TOPIC)
	    (client-display (string-append "Topic on " second-middle " is: "
					   trailing "\n")))

	   ;; Topic change.
	   ((string=? command "TOPIC")
	    (client-display data) (newline)
	    (display (string-append user " changed the topic on " first-middle
				    " to " trailing "\n")))

	   ;; MOTD
	   ((or (string=? command RPL_MOTDSTART)
		(string=? command RPL_MOTD)
		(string=? command RPL_ENDOFMOTD))
	    (client-display (string-append "[motd] " trailing "\n")))

	   ((client-display (string-append "Unknown command: `"
					   prefix "' `"
					   command "' `"
					   first-middle "' `"
					   second-middle "' `"
					   third-middle "' `"
					   middle "' `"
					   trailing "'\n"))))))))

;; Ping our server after having not received a message for this many
;; seconds.
(define max-silence 200)

;; Enter the server loop.  Read a line from SOCKET and pass it to
;; parse-packet.  If parse-packet returns false or EOF of file on SOCKET,
;; is reached, return.  Otherwise, loop.
(define-public (irc-loop socket)
  (if (eqv? (car (select (list (port->fdes socket)) '() '() max-silence))
	    '())
      (if (> irc-outstanding-server-pings 0)
	  (begin (display "Ping timeout\n")
		 (if (not (port-closed? socket))
		     close-port socket))
	  (and (irc-ping-server socket)
	       (irc-loop socket)))
      (let ((data 
	     (catch
	      #t
	      (lambda () (read-line socket))
	      (lambda args #f))))
	(if data
	    (or (eof-object? data)
		(not (parse-packet (substring data 0 (1- (string-length data)))
				   socket))
		(irc-loop socket))
	    (client-display "Socket read error.\n")))))
